/*
 * Copyright (c) 2006-2008 Amit Singh/Google Inc.
 * Copyright (c) 2012 Tuxera Inc.
 * Copyright (c) 2015-2020 Benjamin Fleischer
 * All rights reserved.
 */

#include "fuse_kludges.h"

#if M_OSXFUSE_ENABLE_EXCHANGE

#include <libkern/version.h>
#include <string.h>

/*
 * offsetof(struct vnode, v_name)
 *
 *     10.9  -> 176
 *     10.12 -> 184
 *
 * offsetof(struct vnode, v_parent)
 *
 *     10.9  -> 184
 *     10.12 -> 192
 */

struct fuse_kludge_vnode_13
{
    char dummy[176];
    char *v_name;
    vnode_t v_parent;
} __attribute__ ((packed));

struct fuse_kludge_vnode_16
{
    char dummy[184];
    char *v_name;
    vnode_t v_parent;
} __attribute__ ((packed));

__private_extern__
void
fuse_kludge_exchange(vnode_t vp1, vnode_t vp2)
{
    if (version_major > 15) {
        char *tmp_v_name = ((struct fuse_kludge_vnode_16 *)vp1)->v_name;
        ((struct fuse_kludge_vnode_16 *)vp1)->v_name =
        ((struct fuse_kludge_vnode_16 *)vp2)->v_name;
        ((struct fuse_kludge_vnode_16 *)vp2)->v_name = tmp_v_name;

        vnode_t tmp_v_parent = ((struct fuse_kludge_vnode_16 *)vp1)->v_parent;
        ((struct fuse_kludge_vnode_16 *)vp1)->v_parent =
        ((struct fuse_kludge_vnode_16 *)vp2)->v_parent;
        ((struct fuse_kludge_vnode_16 *)vp2)->v_parent = tmp_v_parent;

    } else {
        char *tmp_v_name = ((struct fuse_kludge_vnode_13 *)vp1)->v_name;
        ((struct fuse_kludge_vnode_13 *)vp1)->v_name =
        ((struct fuse_kludge_vnode_13 *)vp2)->v_name;
        ((struct fuse_kludge_vnode_13 *)vp2)->v_name = tmp_v_name;

        vnode_t tmp_v_parent = ((struct fuse_kludge_vnode_13 *)vp1)->v_parent;
        ((struct fuse_kludge_vnode_13 *)vp1)->v_parent =
        ((struct fuse_kludge_vnode_13 *)vp2)->v_parent;
        ((struct fuse_kludge_vnode_13 *)vp2)->v_parent = tmp_v_parent;
    }
}

#endif
