//
//  MounterCommon.c
//  KextLoader
//
//  Created by Benjamin Fleischer on 02.12.12.
//  Copyright (c) 2012-2019 Benjamin Fleischer. All rights reserved.
//

#include <AvailabilityMacros.h>

#include <CoreFoundation/CoreFoundation.h>

#include "MounterCommon.h"

/*
 * This array must be exactly parallel to the kHelperCommandProcs array
 * in "Helper.c".
 */

const BASCommandSpec kMounterCommandSet[] = {
    {
        kGetVersionCommand,         // commandName
        NULL,                       // rightName
        NULL,                       // rightDefaultRule    -- by default, anyone can acquire this right
        NULL,                       // rightDescriptionKey -- key for custom prompt in "AuthorizationPrompts.strings
        NULL                        // userData
	},
    {
        kLoadKextCommand,           // commandName
        kLoadKextRightName,         // rightName
        "allow",                    // rightDefaultRule    -- by default, anyone can acquire this right
        NULL,                       // rightDescriptionKey -- key for custom prompt in "AuthorizationPrompts.strings
        NULL                        // userData
	},
    {
        kUnloadKextCommand,         // commandName
        kUnloadKextRightName,       // rightName
        "allow",                    // rightDefaultRule    -- by default, anyone can acquire this right
        NULL,                       // rightDescriptionKey -- key for custom prompt in "AuthorizationPrompts.strings
        NULL                        // userData
	},
    {
        kSetSysctlsCommand,         // commandName
        kSetSysctlsRightName,       // rightName
        "allow",                    // rightDefaultRule    -- by default, anyone can acquire this right
        NULL,                       // rightDescriptionKey -- key for custom prompt in "AuthorizationPrompts.strings
        NULL                        // userData
    },
    {
        kCreateMountPathCommand,    // commandName
        kCreateMountPathRightName,  // rightName
        "allow",                    // rightDefaultRule    -- by default, anyone can acquire this right
        NULL,                       // rightDescriptionKey -- key for custom prompt in "AuthorizationPrompts.strings
        NULL                        // userData
    },
    {
        NULL,                       // the array is NULL terminated
        NULL,
        NULL,
        NULL,
        NULL
	}
};

#define KLVersionComponentGetInt(components, count, index) \
    ((index) < (count) ? CFStringGetIntValue(CFArrayGetValueAtIndex((components), (index))) : 0)

Boolean GetSystemVersion(int *systemVersionMajor, int *systemVersionMinor, int *systemVersionBugfix)
{
    Boolean retval = true;

    CFURLRef fileURL;
    CFReadStreamRef fileStream;

    CFPropertyListFormat propertyListFormat;
    CFPropertyListRef propertyList = NULL;

    CFStringRef productVersion;
    CFArrayRef productVersionComponents = NULL;
    CFIndex count;

    fileURL = CFURLCreateWithFileSystemPath(kCFAllocatorDefault,
                                            CFSTR("/System/Library/CoreServices/SystemVersion.plist"),
                                            kCFURLPOSIXPathStyle, false);
    fileStream = CFReadStreamCreateWithFile(kCFAllocatorDefault, fileURL);
    CFRelease(fileURL);

    if (!fileStream) {
        goto out;
    }

    if (!CFReadStreamOpen(fileStream)) {
        CFRelease(fileStream);
        goto out;
    }

    propertyList = CFPropertyListCreateWithStream(kCFAllocatorDefault,
                                                  fileStream,
                                                  0,
                                                  kCFPropertyListImmutable,
                                                  &propertyListFormat,
                                                  NULL);

    CFReadStreamClose(fileStream);
    CFRelease(fileStream);

    if (!propertyList) {
        retval = false;
        goto out;
    }

    // Get value of property ProductVersion
    productVersion = CFDictionaryGetValue(propertyList, CFSTR("ProductVersion"));
    if (!productVersion) {
        retval = false;
        goto out;
    }

    productVersionComponents = CFStringCreateArrayBySeparatingStrings(kCFAllocatorDefault,
                                                                      productVersion,
                                                                      CFSTR("."));
    if (!productVersionComponents) {
        retval = false;
        goto out;
    }
    count = CFArrayGetCount(productVersionComponents);

    if (systemVersionMajor) {
        *systemVersionMajor = KLVersionComponentGetInt(productVersionComponents, count, 0);
    }
    if (systemVersionMinor) {
        *systemVersionMinor = KLVersionComponentGetInt(productVersionComponents, count, 1);
    }
    if (systemVersionBugfix) {
        *systemVersionBugfix = KLVersionComponentGetInt(productVersionComponents, count, 2);
    }

out:
    if (propertyList) {
        CFRelease(propertyList);
    }
    if (productVersionComponents) {
        CFRelease(productVersionComponents);
    }

    return retval;
}
