/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2010, 2016 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXPRINTER, NX protocol compression and extensions to this software     */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

//#include "Win64Fix.h"
#include <windows.h>
#include "debug.h"

#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

int c99_vsnprintf(char* str, size_t size, const char* format, va_list ap)
{
  int count = -1;

  if (size != 0)
  count = _vsnprintf_s(str, size, _TRUNCATE, format, ap);
  if (count == -1)
  count = _vscprintf(format, ap);

  return count;
}

int c99_snprintf(char* str, size_t size, const char* format, ...)
{
  int count;
  va_list ap;

  va_start(ap, format);
  count = c99_vsnprintf(str, size, format, ap);
  va_end(ap);

  return count;
}

#define snprintf c99_snprintf
#define vsnprintf c99_vsnprintf

static char DbgLogFile[MAX_PATH] = {0};

static int DbgLogLevel = 4;

__declspec ( dllexport ) fuseLogFd = -1;

void DbgInit(const char *fname, int level)
{
  strcpy(DbgLogFile, fname);

  DbgLogLevel = level;
}

void DbgMsg(int level, const char *fmt, ...)
{
  va_list ap;

//  static int logfd = -1;

  char timeMsg[4096] = {0};

  SYSTEMTIME st;

  char msg[4096];

  char timeStr[64];


  if (level > DbgLogLevel)
  {
    return;
  }

  va_start(ap, fmt);

  if (fuseLogFd == -1 && DbgLogFile[0])
  {
    fuseLogFd = _open(DbgLogFile, _O_APPEND | _O_CREAT | _O_RDWR, _S_IWRITE);
  }

  if (fuseLogFd < 0)
  {
    fuseLogFd = 2;
  }

//  SYSTEMTIME st;

//  char msg[4096];

//  char timeStr[64];

  GetLocalTime(&st);

  snprintf(timeStr, sizeof(timeStr), "%02d:%02d:%02d %03d",
                st.wHour, st.wMinute, st.wSecond, st.wMilliseconds);

  vsnprintf(msg, sizeof(msg), fmt, ap);

  snprintf(timeMsg, sizeof(timeMsg), "[%d][%d] %s %s", (int) GetCurrentProcessId(),
               (int) GetCurrentThreadId(), timeStr, msg);

  _write(fuseLogFd, timeMsg, strlen(timeMsg));

  va_end(ap);
}
