/*

Copyright (c) 2007, 2008 Hiroki Asakawa asakaw@gmail.com

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <sddl.h>
#include <string.h>
#include "mount.h"
#include "public.h"
#include "debug.h"

static HANDLE                g_EventControl = NULL;
static SERVICE_STATUS        g_ServiceStatus;
static SERVICE_STATUS_HANDLE g_StatusHandle = NULL;

static HANDLE	g_EventLog = NULL;

static HANDLE GarbageCollectorStop =NULL;

static CRITICAL_SECTION	g_CriticalSection;
static LIST_ENTRY		g_MountList;

/*
 * NX FIXME.
 */

#ifdef NX_FIXME

#include <aclapi.h>
#include <shlwapi.h>

#define MAX_SESSIONS 1024

static char StateFilePath[MAX_PATH] = {0};
static char StateFileDir[MAX_PATH] = {0};

static HANDLE GarbageCollectorThread = NULL;
static HANDLE GarbageCollectorResetEvent = NULL;

static DWORD GarbageCollectorHandler(PVOID);

static int GarbageCollectorInit();

static void GarbageCollectorCleanUp();

const char STATE_FILE_RIGHTS[] =

  //
  // (D)esire access list : (P)rotected
  //

  "D:P"

  //
  // (A)dd ;; (G)eneric (A)ll rights ;;; to (SY)stem.
  //

  "(A;;GA;;;SY)"

  //
  // (D)eny ;; (G)eneric (A)ll rights;;; to everyone.
  //

  "(D;;GA;;;WD)";

int SaveState();
int LoadState();

void CleanUp();

void RemoveMountEntry(PMOUNT_ENTRY MountEntry);

int SendReleaseIrp(LPCWSTR deviceName);

int SetObjectRights(const char *objName, const char *rights, int inherit);

#endif NX_FIXME /* NX_FIXME */

/*
 * END OF NX_FIXME.
 */

PMOUNT_ENTRY
InsertMountEntry(PDOKAN_CONTROL DokanControl)
{
	PMOUNT_ENTRY	mountEntry;
	mountEntry = malloc(sizeof(MOUNT_ENTRY));

  DbgPrintW(L"InsertMountEntry : Adding [%s] |-> [%s] pair...\n",
                DokanControl -> MountPoint, DokanControl -> DeviceName);

	if (mountEntry == NULL) {
		DbgPrintW(L"InsertMountEntry malloc failed\n");
		return NULL;
	}
	ZeroMemory(mountEntry, sizeof(MOUNT_ENTRY));
	CopyMemory(&mountEntry->MountControl, DokanControl, sizeof(DOKAN_CONTROL));
	InitializeListHead(&mountEntry->ListEntry);

	EnterCriticalSection(&g_CriticalSection);
	InsertTailList(&g_MountList, &mountEntry->ListEntry);

  /*
   * NX FIXME.
   */

  #ifdef NX_FIXME

  /*
   * Save current mount state to file.
   */

  if (SaveState() != 0)
  {
    RemoveMountEntry(mountEntry);

    return NULL;
  }

  #endif /* NX_FIXME */

  /*
   * END OF NX FIXME.
   */

	LeaveCriticalSection(&g_CriticalSection);

	return mountEntry;
}

VOID
RemoveMountEntry(PMOUNT_ENTRY MountEntry)
{
	EnterCriticalSection(&g_CriticalSection);
	RemoveEntryList(&MountEntry->ListEntry);

  #ifdef NX_FIXME

  /*
   * Save current mount state to file.
   */

  SaveState();

  if (RemoveDirectoryW(MountEntry -> MountControl.MountPoint) == FALSE)
  {
    DbgPrintW(L"WARNING: Cannot to remove mount point [%s].\n"
                  L"Error code is : %d.\n", MountEntry -> MountControl.MountPoint,
                      GetLastError());
  }

  #endif /* NX_FIXME */

  /*
   * END OF NX FIXME.
   */

	LeaveCriticalSection(&g_CriticalSection);

	free(MountEntry);
}

PMOUNT_ENTRY
FindMountEntry(PDOKAN_CONTROL	DokanControl)
{
	PLIST_ENTRY		listEntry;
	PMOUNT_ENTRY	mountEntry;
	BOOL			useMountPoint = wcslen(DokanControl->MountPoint) > 0;
	BOOL			found = FALSE;

	if (!useMountPoint && wcslen(DokanControl->DeviceName) == 0) {
		return NULL;
	}

	EnterCriticalSection(&g_CriticalSection);

    for (listEntry = g_MountList.Flink; listEntry != &g_MountList; listEntry = listEntry->Flink) {
		mountEntry = CONTAINING_RECORD(listEntry, MOUNT_ENTRY, ListEntry);
		if (useMountPoint) {
			if (wcscmp(DokanControl->MountPoint, mountEntry->MountControl.MountPoint) == 0) {
				found = TRUE;
				break;
			}
		} else {
			if (wcscmp(DokanControl->DeviceName, mountEntry->MountControl.DeviceName) == 0) {
				found = TRUE;
				break;
			}
		}
	}

	LeaveCriticalSection(&g_CriticalSection);

	if (found) {
		DbgPrintW(L"FindMountEntry %s -> %s\n",
			mountEntry->MountControl.MountPoint, mountEntry->MountControl.DeviceName);
		return mountEntry;
	} else {
		return NULL;
	}
}

VOID
DokanControlFind(PDOKAN_CONTROL Control)
{
	PLIST_ENTRY		listEntry;
	PMOUNT_ENTRY	mountEntry;

	mountEntry = FindMountEntry(Control);
	if (mountEntry == NULL) {
		Control->Status = DOKAN_CONTROL_FAIL;
	} else {
		wcscpy_s(Control->DeviceName, sizeof(Control->DeviceName) / sizeof(WCHAR),
				mountEntry->MountControl.DeviceName);
		wcscpy_s(Control->MountPoint, sizeof(Control->MountPoint) / sizeof(WCHAR),
				mountEntry->MountControl.MountPoint);
		Control->Status = DOKAN_CONTROL_SUCCESS;
	}
}

VOID
DokanControlList(PDOKAN_CONTROL Control)
{
	PLIST_ENTRY		listEntry;
	PMOUNT_ENTRY	mountEntry;
	ULONG			index = 0;

	EnterCriticalSection(&g_CriticalSection);
	Control->Status = DOKAN_CONTROL_FAIL;

	for (listEntry = g_MountList.Flink;
		listEntry != &g_MountList;
		listEntry = listEntry->Flink) {
		mountEntry = CONTAINING_RECORD(listEntry, MOUNT_ENTRY, ListEntry);
		if (Control->Option == index++) {
			wcscpy_s(Control->DeviceName, sizeof(Control->DeviceName) / sizeof(WCHAR),
					mountEntry->MountControl.DeviceName);
			wcscpy_s(Control->MountPoint, sizeof(Control->MountPoint) / sizeof(WCHAR),
					mountEntry->MountControl.MountPoint);
			Control->Status = DOKAN_CONTROL_SUCCESS;
			break;
		}
	}
	LeaveCriticalSection(&g_CriticalSection);
}

static VOID DokanControl(PDOKAN_CONTROL Control)
{
	PMOUNT_ENTRY	mountEntry;
	ULONG	index = 0;
	DWORD written = 0;

	Control->Status = DOKAN_CONTROL_FAIL;

	switch (Control->Type)
	{
	case DOKAN_CONTROL_MOUNT:

		DbgPrintW(L"DokanControl Mount\n");

   /*
    * NX FIXME: If save state inside InserMountEntry() fails
    *           then do not add volume.
    */

    #ifdef NX_FIXME

    Control -> Status   = DOKAN_CONTROL_FAIL;
    Control -> StatusEx = NXFSD_STATUS_SUCCESS;

    if (InsertMountEntry(Control) == NULL)
    {
      Control -> StatusEx = NXFSD_STATUS_DB_ERROR;

      break;
    }

    if (DokanControlMount(Control -> MountPoint, Control -> DeviceName) == FALSE)
    {
      break;
    }

    Control -> Status = DOKAN_CONTROL_SUCCESS;

    /*
     * Tell garbage collector that processes list chanded.
     */

    SetEvent(GarbageCollectorResetEvent);

    /*
     * END OF NX FIXME.
     */

    #else /* NX_FIXME */

		if (DokanControlMount(Control->MountPoint, Control->DeviceName)) {
			Control->Status = DOKAN_CONTROL_SUCCESS;
			InsertMountEntry(Control);
		} else {
			Control->Status = DOKAN_CONTROL_FAIL;
		}

    #endif /* else NX_FIXME */

		break;

	case DOKAN_CONTROL_UNMOUNT:

		DbgPrintW(L"DokanControl Unmount\n");

		mountEntry = FindMountEntry(Control);

     /*
      * NX FIXME:
      * - Do not use unmount force (i.e. stop if mount point not in db).
      *
      * - Don't stop when mount point directory not found on
      *   drive - we still need to delete volume object in driver.
      *
      * - Send IOCTL to driver (originaly it is done in client
      *   process via dokan.dll.
      *   Reason: When client process fails we can't remove volume object
      *           causing resource lack, when service can restore volume
      *           object from DB even after crash.
      */

      #ifdef NX_FIXME

      Control -> Status = DOKAN_CONTROL_FAIL;

      /*
       * Not NX mount poins or database corrupted.
       */

      if (mountEntry == NULL)
      {
        Control -> StatusEx = NXFSD_STATUS_NOT_NX;
        Control -> Status   = DOKAN_CONTROL_FAIL;

        break;
      }

      mountEntry -> MountControl.Status = DOKAN_CONTROL_FAIL;

      /*
       * Clear reparse point.
       */

      if (DokanControlUnmount(mountEntry -> MountControl.MountPoint))
      {
        Control -> Status = DOKAN_CONTROL_SUCCESS;
      }

      /*
       * Send unount IOCTL to driver.
       */

      SendReleaseIrp(mountEntry -> MountControl.DeviceName);

      /*
       * Remove entry from db.
       */

      RemoveMountEntry(mountEntry);

      /*
       * Tell garbage collector, that processes list chanded.
       */

      SetEvent(GarbageCollectorResetEvent);

      /*
       * END OF NX FIXME.
       */

      #else /* NX_FIXME */

		if (mountEntry == NULL) {
			if (Control->Option == DOKAN_CONTROL_OPTION_FORCE_UNMOUNT &&
				DokanControlUnmount(Control->MountPoint)) {
				Control->Status = DOKAN_CONTROL_SUCCESS;
				break;
			}
			Control->Status = DOKAN_CONTROL_FAIL;
			break;
		}

		if (DokanControlUnmount(mountEntry->MountControl.MountPoint)) {
			Control->Status = DOKAN_CONTROL_SUCCESS;
			if (wcslen(Control->DeviceName) == 0) {
				wcscpy_s(Control->DeviceName, sizeof(Control->DeviceName) / sizeof(WCHAR),
						mountEntry->MountControl.DeviceName);
			}
			RemoveMountEntry(mountEntry);
		} else {
			mountEntry->MountControl.Status = DOKAN_CONTROL_FAIL;
			Control->Status = DOKAN_CONTROL_FAIL;
		}

      #endif /* else NX_FIXME */

		break;

	case DOKAN_CONTROL_CHECK:
		{
			DbgPrint("DokanControl Check\n");
			Control->Status = 0;
		}
		break;

	case DOKAN_CONTROL_FIND:
		{
			DbgPrintW(L"DokanControl Find\n");
			DokanControlFind(Control);
		}
		break;

	case DOKAN_CONTROL_LIST:
		{
			DbgPrintW(L"DokanControl List\n");
			DokanControlList(Control);
		}
		break;

	default:
		DbgPrintW(L"DokanControl UnknownType %u\n", Control->Type);
	}

	return;
}

static VOID BuildSecurityAttributes(PSECURITY_ATTRIBUTES SecurityAttributes)
{
	LPTSTR sd = L"D:P(A;;GA;;;SY)(A;;GRGWGX;;;BA)(A;;GRGW;;;WD)(A;;GR;;;RC)";

	ZeroMemory(SecurityAttributes, sizeof(SECURITY_ATTRIBUTES));

	ConvertStringSecurityDescriptorToSecurityDescriptor(
		sd,
		SDDL_REVISION_1,
		&SecurityAttributes->lpSecurityDescriptor,
		NULL);

	SecurityAttributes->nLength = sizeof(SECURITY_ATTRIBUTES);
    SecurityAttributes->bInheritHandle = TRUE;
}

/*
 *  We need notification from garbage collector if something failed.
 */

HANDLE GarbageCollectorStopEvent;

DWORD WINAPI FuseServiceMain(LPVOID fuseEvents)
{
  DWORD eventNo;
  HANDLE pipe, device;
  HANDLE eventConnect, eventUnmount;
  HANDLE eventArray[4];
  DOKAN_CONTROL control, unmount;
  OVERLAPPED ov, driver;
  ULONG returnedBytes;
  EVENT_CONTEXT eventContext;
  SECURITY_ATTRIBUTES sa;
  HANDLE FuseReadyEvent, FuseStopEvent, FuseStopRequest;
  DWORD errorCode;
  DWORD exitCode = 0;
  int err;
  int retries = 0;

  FuseStopEvent = *((HANDLE *) fuseEvents);
  FuseReadyEvent = *(((HANDLE *) fuseEvents) + 1);
  FuseStopRequest = *(((HANDLE *) fuseEvents) + 2);

  #if _MSC_VER < 1300
  InitializeCriticalSection(&g_CriticalSection);
  #else
  InitializeCriticalSectionAndSpinCount(&g_CriticalSection, 0x80000400);
  #endif

  InitializeListHead(&g_MountList);

  DBG_MSG("FuseServiceMain: Initialising fuse thread.\n");

  /*
   * NX FIXME:
   * - Restore mount state from file if exist.
   * - Do clean up if any entries remains from last execution.
   * - Set up thread to catch dead client processes.
   */

  LoadState();

  CleanUp();

  GarbageCollectorInit();

  if (GarbageCollectorStopEvent == NULL)
  {
    DBG_MSG("FuseServiceMain: ERROR! Failed to initialize"
                " garbage collector.\n");

    SetEvent(FuseStopEvent);

    DeleteCriticalSection(&g_CriticalSection);

    return -1;
  }

  /*
   * END OF NX FIXME.
   */

  BuildSecurityAttributes(&sa);

  DBG_MSG("FuseServiceMain: Creating dokan control pipe with old security atributes.\n");

  pipe = CreateNamedPipe(DOKAN_CONTROL_PIPE,
                            PIPE_ACCESS_DUPLEX | FILE_FLAG_OVERLAPPED,
                               PIPE_TYPE_MESSAGE | PIPE_READMODE_MESSAGE |
                                   PIPE_WAIT, 1, sizeof(control),
                                      sizeof(control), 1000, &sa);

  if (pipe == INVALID_HANDLE_VALUE)
  {
    DBG_MSG("FuseServiceMain: WARNING! Failed to create pipe.\n");

    err = GetLastError();

    DBG_MSG("FuseServiceMain: Error is %d.\n", err);

    if (err == ERROR_PIPE_BUSY)
    {
      /*
       *  Resources can be temporary unavailable, we
       *  should retry.
       */

      while (retries < 20)
      {
        if (WaitNamedPipe(DOKAN_CONTROL_PIPE, 2000) == 0)
        {
          DBG_MSG("FuseServiceMain: WARNING! Waiting for instance failed.\n");
        }
        else
        {
          pipe = CreateNamedPipe(DOKAN_CONTROL_PIPE, PIPE_ACCESS_DUPLEX | \
                                     FILE_FLAG_OVERLAPPED, PIPE_TYPE_MESSAGE \
                                         | PIPE_READMODE_MESSAGE | PIPE_WAIT,
                                             1, sizeof(control), sizeof(control),
                                                 1000, NULL);

          if (pipe == INVALID_HANDLE_VALUE)
          {
            DBG_MSG("FuseServiceMain: WARNING! Retry of pipe creation failed.\n");

            DBG_MSG("FuseServiceMain: Error is %d.\n", GetLastError());
          }
          else
          {
            break;
          }
        }

        retries++;
      }

      if (pipe == INVALID_HANDLE_VALUE)
      {
        DBG_MSG("FuseServiceMain: ERROR! Failed to create pipe.\n");

        SetEvent(FuseStopEvent);

        GarbageCollectorCleanUp();

        CloseHandle(GarbageCollectorStopEvent);

        DeleteCriticalSection(&g_CriticalSection);

        return -1;
      }
    }
    else
    {
      DBG_MSG("FuseServiceMain: ERROR! Failed to create pipe.\n");

      SetEvent(FuseStopEvent);

      GarbageCollectorCleanUp();

      CloseHandle(GarbageCollectorStopEvent);

      DeleteCriticalSection(&g_CriticalSection);

      return -1;
    }
  }

  device = CreateFile(DOKAN_GLOBAL_DEVICE_NAME, GENERIC_READ |
                          GENERIC_WRITE, FILE_SHARE_READ |
                              FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
                                  FILE_FLAG_OVERLAPPED, NULL);

  if (device == INVALID_HANDLE_VALUE)
  {
    DBG_MSG("FuseServiceMain: ERROR! Failed to open nxfs device.\n");

    DBG_MSG("FuseServiceMain: Error is %d.\n", GetLastError());

    SetEvent(FuseStopEvent);

    CloseHandle(pipe);

    GarbageCollectorCleanUp();

    DbgPrint("FuseServiceMain: Closing GarbageCollector stop event.\n");

    CloseHandle(GarbageCollectorStopEvent);

    DeleteCriticalSection(&g_CriticalSection);

    return -1;
  }

  eventConnect = CreateEvent(NULL, FALSE, FALSE, NULL);
  eventUnmount = CreateEvent(NULL, FALSE, FALSE, NULL);

  if (eventConnect == NULL || eventUnmount == NULL)
  {
    DBG_MSG("FuseServiceMain: ERROR! Can't create events.\n");

    DBG_MSG("FuseServiceMain: Error is %d.\n", GetLastError());

    SetEvent(FuseStopEvent);

    CloseHandle(pipe);

    if (eventConnect != NULL)
    {
      CloseHandle(eventConnect);
    }

    if (eventUnmount != NULL)
    {
      CloseHandle(eventUnmount);
    }

    GarbageCollectorCleanUp();

    CloseHandle(GarbageCollectorStopEvent);

    DeleteCriticalSection(&g_CriticalSection);

    return -1;
  }

  for (;;)
  {
    ZeroMemory(&ov, sizeof(OVERLAPPED));
    ZeroMemory(&driver, sizeof(OVERLAPPED));
    ZeroMemory(&eventContext, sizeof(EVENT_CONTEXT));

    ov.hEvent = eventConnect;
    driver.hEvent = eventUnmount;

    if (ConnectNamedPipe(pipe, &ov) == 0)
    {
      errorCode = GetLastError();

      if (errorCode != ERROR_IO_PENDING && errorCode != ERROR_PIPE_CONNECTED)
      {
        DBG_MSG("FuseServiceMain: ERROR! Can't connect pipe.\n");

        DBG_MSG("FuseServiceMain: Error is %d.\n", errorCode);

        SetEvent(FuseStopEvent);

        CloseHandle(pipe);

        CloseHandle(eventConnect);

        CloseHandle(eventUnmount);

        GarbageCollectorCleanUp();

        CloseHandle(GarbageCollectorStopEvent);

        DeleteCriticalSection(&g_CriticalSection);

        return -1;
      }
    }

    if (!DeviceIoControl(device, IOCTL_SERVICE_WAIT, NULL, 0,
                             &eventContext, sizeof(EVENT_CONTEXT),
                                 NULL, &driver))
    {
      DWORD error = GetLastError();

      if (error != 997)
      {
        DBG_MSG("FuseServiceMain: ERROR! Can't send control\
                     code to nxfs device.\n");

        DBG_MSG("FuseServiceMain: Error is: %d.\n", error);

        SetEvent(FuseStopEvent);

        CloseHandle(pipe);

        CloseHandle(eventConnect);

        CloseHandle(eventUnmount);

        GarbageCollectorCleanUp();

        CloseHandle(GarbageCollectorStopEvent);

        DeleteCriticalSection(&g_CriticalSection);

        return -1;
      }
    }

    eventArray[0] = eventConnect;
    eventArray[1] = eventUnmount;
    eventArray[2] = FuseStopRequest;
    eventArray[3] = GarbageCollectorStopEvent;

    //
    // Notify main thread that initialization
    // is completed.
    //

    SetEvent(FuseReadyEvent);

    DBG_MSG("FuseServiceMain: Waiting for disk events.\n");

    eventNo = WaitForMultipleObjects(4, eventArray, FALSE, INFINITE)\
              - WAIT_OBJECT_0;

    DBG_MSG("FuseServiceMain: Received event number: %d.\n", (int) eventNo);

    if (eventNo == 0)
    {
      DWORD result = 0;

      DBG_MSG("FuseServiceMain: Processing connect event.\n");

      ZeroMemory(&control, sizeof(control));

      if (ReadFile(pipe, &control, sizeof(control), &result, NULL))
      {
        DokanControl(&control);
        WriteFile(pipe, &control, sizeof(control), &result, NULL);
      }

      FlushFileBuffers(pipe);

      DisconnectNamedPipe(pipe);
    }
    else if (eventNo == 1)
    {
      if (GetOverlappedResult(device, &driver, &returnedBytes, FALSE))
      {
        if (returnedBytes == sizeof(EVENT_CONTEXT))
        {
          DBG_MSG("FuseServiceMain: Processing unmount event.\n");

          ZeroMemory(&unmount, sizeof(DOKAN_CONTROL));

          unmount.Type = DOKAN_CONTROL_UNMOUNT;

          wcscpy_s(unmount.DeviceName, sizeof(unmount.DeviceName) /
                       sizeof(WCHAR), eventContext.Unmount.DeviceName);

          DokanControl(&unmount);
        }
        else
        {
          DBG_MSG("FuseServiceMain: ERROR! Unmount failed.\n");

          exitCode = -1;

          break;
        }
      }
    }
    else if (eventNo == 2)
    {
      DBG_MSG("FuseServiceMain: Received stop request from main thread.\n");

      break;
    }
    else if (eventNo == 3)
    {
      DBG_MSG("FuseServiceMain: Received stop event from "
                  "garbage collector thread.\n");

      CloseHandle(GarbageCollectorStopEvent);

      break;
    }
    else
    {
      break;
    }
  }

  if (pipe != NULL)
  {
    DisconnectNamedPipe(pipe);

    CloseHandle(pipe);

    pipe = NULL;
  }

  if (eventConnect != NULL)
  {
    CloseHandle(eventConnect);
  }

  if (device != NULL)
  {
    CloseHandle(device);
  }

  if (eventUnmount != NULL)
  {
    CloseHandle(eventUnmount);
  }

  GarbageCollectorCleanUp();

  CloseHandle(GarbageCollectorStopEvent);

  DeleteCriticalSection(&g_CriticalSection);

  SetEvent(FuseStopEvent);

  return exitCode;
}

/*
 * NX FIXME.
 */

#ifdef NX_FIXME

#define FAIL(X) if (X) goto fail

#define FAILEX(X, ...) if (X) {DbgPrint(__VA_ARGS__); goto fail;}

#define Not(X) (!(X))

//
// Transform relative (to where nxfsd stored) path into full path.
// E.g. it transforms "/somedir" into "<where is nxfsd.exe>/somedir".
// We use it becouse daemon (service) doesn't start in path, where nxdevice.exe
// stored until we don't set current directory manually.
//
// WARNING: If passed path is already absolute it returns the same string.
//
// normPath - full, generated path (OUT).
// pathSize - size of path buffer in bytes (IN).
// relative - relative path postfix to add to base path (IN).
//
// RETURNS: 0 if OK.
//

int ExpandRelativePath(char *normPath, int pathSize, char *relative)
{
  int exitCode = -1;

  int lastSlash = 0;

  int i = 0;

  char *path = (char *) malloc(pathSize);

  DbgPrint("-> ExpandRelativePath()...\n");

  FAILEX(path == NULL, "ERROR: Out of memory.\n");

  //
  // Check is it already absolute?
  //

  if (relative && relative[0] && relative[1] == ':')
  {
    DbgPrint("Path [%s] is already absolute.\n", relative);

    strncpy(path, relative, pathSize);
  }
  else
  {
    DbgPrint("Expanding path [%s]...\n", relative);

    //
    // Get full path to nxdevice.exe binary.
    //

    FAIL(GetModuleFileNameA(NULL, path, pathSize) == FALSE);

    //
    // Remove filename from path.
    //

    for (i = 0; path[i] && i < pathSize; i++)
    {
      if (path[i] == '\\' || path[i] == '/')
      {
        lastSlash = i;
      }
    }

    path[lastSlash] = 0;

    //
    // Add relative postfix if any.
    //

    if (relative)
    {
      strncat(path, "\\", pathSize);
      strncat(path, relative, pathSize);
    }
  }

  //
  // Normalize path for output (e.g. remove '..' entries).
  //

  for (i = 0; path[i]; i++)
  {
    if (path[i] == '/')
    {
      path[i] = '\\';
    }
  }

  PathCanonicalizeA(normPath, path);

  exitCode = 0;

  DbgPrint("Path [%s] expanded to [%s].\n", relative, normPath);

  //
  // Error handler.
  //

  fail:

  if (exitCode)
  {
    int err = GetLastError();

    DbgPrint("ERROR: Cannot retrieve path of binary.\n"
                  "Error code is : %d.\n", err);

    if (err != 0)
    {
      exitCode = err;
    }
  }

  if (path != NULL)
  {
    free(path);
  }

  DbgPrint("<- ExpandRelativePath()...\n");

  return exitCode;
}

int GetStateFileDir()
{
  const char *dirName = "nxdevice";

  char tempPath[MAX_PATH] = {0};

  if (StateFileDir[0] == 0)
  {
    if (GetTempPathA(MAX_PATH, tempPath) == 0 ||
            !PathIsDirectoryA(tempPath))
    {
      return -1;
    }

    _snprintf(StateFileDir, MAX_PATH, "%s\\%s", tempPath, dirName);
  }

  return 0;
}

BOOL DirectoryExists(LPCSTR path)
{
  DWORD attr = GetFileAttributesA(path);

  return attr != INVALID_FILE_ATTRIBUTES &&
            attr & FILE_ATTRIBUTE_DIRECTORY;
}

int CreateStateFileDir()
{
  if (GetStateFileDir())
  {
    return -1;
  }

  if (!DirectoryExists(StateFileDir))
  {
    return !CreateDirectoryA(StateFileDir, NULL);
  }

  return 0;
}

/*
 * Retrieve path to state file.
 *
 * RETURNS: 0 if OK.
 */

int GetStateFilePath()
{
  const char *fileName = "fstab";

  if (StateFilePath[0] == 0)
  {
    if (GetStateFileDir())
    {
      return -1;
    }

    _snprintf(StateFilePath, MAX_PATH, "%s\\%s", StateFileDir, fileName);
  }

  DbgPrint("Set state file to [%s].\n", StateFilePath);

  return 0;
}


/*
 * Restore mount state from file.
 *
 * RETURNS: 0 if OK.
 */

int LoadState()
{
  int exitCode = -1;

  DOKAN_CONTROL dokanControl;

  PMOUNT_ENTRY mountEntry = NULL;

  FILE *f = NULL;

  DbgPrint("-> LoadState()...\n");

  FAIL(GetStateFilePath() != 0);

  EnterCriticalSection(&g_CriticalSection);

  /*
   * Open state file.
   */

  f = fopen(StateFilePath, "rb");

  FAIL(f == NULL);

  /*
   * Load state from file.
   */

  while(feof(f) == 0)
  {
    /*
     * Load next record from file.
     */

    if (fread(&dokanControl, sizeof(DOKAN_CONTROL), 1, f) == 1)
    {
      /*
       * Allocate new list's element.
       */

      mountEntry = malloc(sizeof(MOUNT_ENTRY));

      FAIL(mountEntry == NULL);

      /*
       * Put readed record into list element.
       */

      ZeroMemory(mountEntry, sizeof(MOUNT_ENTRY));

      CopyMemory(&mountEntry -> MountControl, &dokanControl, sizeof(DOKAN_CONTROL));

      /*
       * Put new element into list.
       */

      InsertTailList(&g_MountList, &mountEntry -> ListEntry);

      /*
       * Log element.
       */

      DbgPrintW(L"Restored mount point [%s] -> [%s]\n",
                    mountEntry -> MountControl.MountPoint,
                        mountEntry -> MountControl.DeviceName);

    }
  }

  /*
   * Error handler.
   */

  exitCode = 0;

  fail:

  if (exitCode)
  {
    DbgPrint("LoadState : ERROR: Cannot open state file.\n"
                 "Error code is : %d.\n", GetLastError());
  }

  if (f != NULL)
  {
    fclose(f);
  }

  LeaveCriticalSection(&g_CriticalSection);

  DbgPrint("<- LoadState()...\n");

  return exitCode;
}

/*
 * Save mount state to file.
 *
 * RETURNS: 0 if OK.
 */

int SaveState()
{
  int exitCode = -1;

  PLIST_ENTRY listEntry;

  PMOUNT_ENTRY mountEntry;

  char tempPath[MAX_PATH] = {0};
  char tempName[MAX_PATH] = {0};

  FILE *f = NULL;


  DbgPrint("-> SaveState()...\n");

  FAIL(CreateStateFileDir() != 0);
  FAIL(GetStateFilePath() != 0);

  EnterCriticalSection(&g_CriticalSection);

  /*
   * FIXME. Use native transacted I/O on Vista/Win7.
   */

  /*
   * Create temp file name.
   */

  FAIL(GetTempPathA(MAX_PATH, tempPath) == 0);

  FAIL(GetTempFileNameA(tempPath, ".nx", 0, tempName) == 0);

  /*
   * Open tmp state file to write.
   * We use tmp file to make atomic write on WinXP.
   */

  DbgPrint("SaveState : Opening temp file [%s]...\n", tempName);

  f = fopen(tempName, "wb+");

  FAIL(f == NULL);

  /*
   * Dump state to tmp file.
   */

  for (listEntry = g_MountList.Flink;
           listEntry != &g_MountList;
               listEntry = listEntry -> Flink)
  {
    mountEntry = CONTAINING_RECORD(listEntry, MOUNT_ENTRY, ListEntry);

    DbgPrintW(L"SaveState : Storing [%s] |-> [%s]...\n",
                 mountEntry -> MountControl.MountPoint,
                     mountEntry -> MountControl.DeviceName);

    fwrite(&(mountEntry -> MountControl), sizeof(DOKAN_CONTROL), 1, f);
  }

  fclose(f);

  /*
   * Replace real state file with tmp.
   *
   */

  DbgPrint("SaveState : Replacing [%s] with [%s]...\n", StateFilePath, tempName);

  if (ReplaceFileA((LPCSTR) StateFilePath,
                       (LPCSTR) tempName,
                           NULL, 0, NULL, NULL) == FALSE)
  {
    /*
     * If state file not exist yet use MoveFileEx.
     */

    FAIL(GetLastError() != ERROR_FILE_NOT_FOUND);

    FAIL(MoveFileExA((LPCSTR) tempName,
                         (LPCSTR) StateFilePath,
                             MOVEFILE_WRITE_THROUGH) == FALSE);
  }

  /*
   * Forbid non-SYSTEM users to change state file (including delete)
   * to make sure anybody don't corrupt database file.
   */

  FAIL(SetObjectRights(StateFilePath, STATE_FILE_RIGHTS, 0));

  /*
   * Clean up.
   */

  exitCode = 0;

  fail:

  if (exitCode)
  {
    DbgPrint("ERROR: Cannot to save state.\n"
                 "Error code is %d.\n", GetLastError());
  }

  LeaveCriticalSection(&g_CriticalSection);

  DbgPrint("<- SaveState()...\n");

  return exitCode;
}

/*
 * Unmount all disks. We call it at service begin and end.
 */

void CleanUp()
{
  PLIST_ENTRY listEntry;

  PMOUNT_ENTRY mountEntry;

  DbgPrint("-> CleanUp()...\n");

  /*
   * Interate all entries in database.
   */

  for (listEntry = g_MountList.Flink;
           listEntry != &g_MountList;
               listEntry = listEntry -> Flink)
  {
    mountEntry = CONTAINING_RECORD(listEntry, MOUNT_ENTRY, ListEntry);

    /*
     * Clear reparse point.
     */

    DbgPrintW(L"CleanUp : Clearing reparse point on [%s]...\n",
                  mountEntry -> MountControl.MountPoint);

    if (DokanControlUnmount(mountEntry -> MountControl.MountPoint) == FALSE)
    {
      DbgPrintW(L"WARNING: Cannot unmount [%s] entry.\n"
                    L"Error code is : [%d].",
                        mountEntry -> MountControl.MountPoint,
                            GetLastError());
    }

    /*
     * Send release IOCTL to driver.
     */

    DbgPrintW(L"CleanUp : Sending release IOCTL for volume device [%s]...\n",
                  mountEntry -> MountControl.DeviceName);

    SendReleaseIrp(mountEntry -> MountControl.DeviceName);

    /*
     * Remove entry from db.
     */

    RemoveMountEntry(mountEntry);
  }

  /*
   * Remove state file to avoid access denien on whole NX uninstall.
   *
   * Reason: We deny non-SYSTEM users to modify DB file to avoid
   * corruption.
   */

  DbgPrint("CleanUp : Removing state file...");

  if (DeleteFileA(StateFilePath) == FALSE)
  {
    DbgPrint("WARNING: Cannot remove state file [%s].\n"
                 "Error code is : %d.\n", StateFilePath, GetLastError());
  }
  else
  {
    RemoveDirectoryA(StateFileDir);
  }

  DbgPrint("<- CleanUp()...\n");
}

/*
 * Send release IOCTL for given volume object.
 * In original dokan, release IOCTL is sent to driver
 * in client DLL, but when client process fail (or simply
 * someone kill him, we can't remove volume and have resource lack).
 *
 * Service stores device [mount_point] |-> [deviceName] pairs in
 * external DB.
 *
 * deviceName - volume name to release (IN).
 *
 * RETURNS: 0 if OK.
 */

int SendReleaseIrp(LPCWSTR deviceName)
{
  HANDLE device;

  BOOL status;

  OVERLAPPED ov = {0};

  DWORD waitResult = 0;

  WCHAR globalName[MAX_PATH] = L"\\\\.";

  int exitCode = -1;

  DbgPrint("-> SendReleaseIrp()...\n");

  /*
   * Create global device name.
   */

  wcscat_s(globalName, MAX_PATH, deviceName);

  /*
   * Open volume device.
   */

  DbgPrintW(L"Opening [%s] device...\n", globalName);

  device = CreateFile(globalName, GENERIC_READ | GENERIC_WRITE,
                          FILE_SHARE_READ | FILE_SHARE_WRITE,
                              NULL, OPEN_EXISTING,
                                  FILE_FLAG_OVERLAPPED,
                                      NULL);

  FAILEX(device == INVALID_HANDLE_VALUE,
             "ERROR: Cannot open volume device.\n");

  /*
   * Prepare overlapped mode.
   */

  ov.hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);

  /*
   * Send release IOCTL to volume device.
   */

  DbgPrintW(L"Sending release to volume device [%s]...\n", deviceName);

  DeviceIoControl(device, IOCTL_EVENT_RELEASE, NULL, 0, NULL, 0, NULL, &ov);

  /*
   * Wait until driver responde or timeout reached.
   */

  waitResult = WaitForSingleObject(ov.hEvent, 10000);

  FAILEX(waitResult != WAIT_OBJECT_0,
             "ERROR: Wait failed with result [%d].\n", GetLastError());

  /*
   * Clean up.
   */

  exitCode = 0;

  fail:

  if (exitCode)
  {
    DbgPrintW(L"ERROR: Cannot send release IOCTL for volume [%s].\n"
                  L"Error code is : %d.\n", deviceName, GetLastError());
  }

  CloseHandle(device);
  CloseHandle(ov.hEvent);

  DbgPrint("<- SendReleaseIrp()...\n");

  return exitCode;
}

/*
 * Set security string for given named object (e.g. file).
 *
 * objName - object name (e.g. path to file). (IN)
 * rights  - security string with access list to set (IN).
 * inherit - New ACL inherits rights from parent if 1 (IN).
 *
 * RETURNS: 0 if OK.
 */

int SetObjectRights(const char *objName, const char *rights, int inherit)
{
  int exitCode = -1;

  ACL *acl = NULL;

  SECURITY_DESCRIPTOR sd = {0};

  PSECURITY_DESCRIPTOR psd = &sd;

  DWORD flags;

  BOOL aclExists  = FALSE;
  BOOL aclDefault = FALSE;

  /*
   * Initialize empty security descriptor.
   */

  FAILEX(InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION) == FALSE,
             "ERROR: Cannot initialize security descriptor.\n");

  /*
   * Set NULL access list for security descriptor.
   */

  FAILEX(SetSecurityDescriptorDacl(&sd, TRUE, NULL, FALSE) == FALSE,
             "ERROR: Cannot set DACL for security descriptor.\n");

  /*
   * Init security descriptor from security string.
   */

  FAILEX(ConvertStringSecurityDescriptorToSecurityDescriptorA(rights, SDDL_REVISION_1,
                                                                  &psd, NULL) == FALSE,
             "ERROR: Cannot process security string.\n");

  /*
   * Retrieve access list from craeted security descriptor.
   */

  FAILEX(GetSecurityDescriptorDacl(psd, &aclExists, &acl, &aclDefault) == FALSE,
             "ERROR: Cannot retrive acl from security descriptor.\n");

  /*
   * Set new access list to named object.
   */

  flags = DACL_SECURITY_INFORMATION;

  if (Not(inherit))
  {
    flags |= PROTECTED_DACL_SECURITY_INFORMATION;
  }

  FAILEX(SetNamedSecurityInfoA((char *) objName, SE_FILE_OBJECT, flags,
                                   NULL, NULL, acl, NULL),
                                       "ERROR: Cannot set access list.\n")

  /*
   * Clean up.
   */

  exitCode = 0;

  fail:

  if (exitCode)
  {
    DbgPrint("ERROR: Cannot set security string [%ls] for [%s].\n"
                 "Error code is : %d.\n", rights, objName, GetLastError());
  }

  if (psd)
  {
    LocalFree(psd);
  }

  return exitCode;
}

/*
 * Garbage collector thread to detect dead client processes.
 */

DWORD GarbageCollectorHandler(PVOID unused)
{
  int exitCode = -1;

  int goOn = 1;

  static HANDLE handles[MAX_SESSIONS + 1];

  static DOKAN_CONTROL handlesData[MAX_SESSIONS];

  PLIST_ENTRY listEntry;

  PMOUNT_ENTRY mountEntry;

  DWORD count = 0;
  DWORD pid   = 0;
  DWORD ret   = 0;
  DWORD id    = 0;

  DbgPrint("-> GarbageCollectorHandler()...\n");

  /*
   * Main loop.
   */

  while(goOn)
  {
    /*
     * Put all client processes to wait.
     */

    EnterCriticalSection(&g_CriticalSection);

    count = 0;

    for (listEntry = g_MountList.Flink;
             listEntry != &g_MountList;
                 listEntry = listEntry -> Flink)
    {
      mountEntry = CONTAINING_RECORD(listEntry, MOUNT_ENTRY, ListEntry);

      pid = mountEntry -> MountControl.clientPid;

      DbgPrint("GarbageCollector : putting pid [%d] to wait list...\n", pid);

      handles[count] = OpenProcess(SYNCHRONIZE, FALSE, pid);

      memcpy(handlesData + count,
                 &mountEntry -> MountControl,
                     sizeof(DOKAN_CONTROL));

      count ++;

      if (count == MAX_SESSIONS)
      {
        #ifdef ERROR
        DBG_MSG("GarbageCollectorHandler: ERROR: Too many sessions.\n");
        #endif

        LeaveCriticalSection(&g_CriticalSection);

        exitCode = 1;

        goto fail;
      }
    }

    LeaveCriticalSection(&g_CriticalSection);

    /*
     * Put reset event to know, when new process arived or
     * sent unmount request directly.
     */

    handles[count]   = GarbageCollectorResetEvent;

    count ++;

    FAILEX(count == MAX_SESSIONS, "ERROR: Too many sessions.\n");

    /*
     * Wait until one of process dead or reset event arrived.
     */

    handles[count] = GarbageCollectorStop;

    ret = WaitForMultipleObjects(count + 1, handles, FALSE, INFINITE);

    /*
     * Check which handle signaled.
     */

    id = ret - WAIT_OBJECT_0;

    FAILEX(id >= MAX_SESSIONS, "ERROR: Wait failed with code [%d].", ret);
    FAILEX(id > count - 1, "ERROR: Wait failed with code [%d].", ret);

    /*
     * Reset event. Only reset thread loop.
     */

    if (handles[id] == GarbageCollectorStop)
    {
      DBG_MSG("GarbageCollector: Received stop event.\n");

      break;
    }

    if (handles[id] == GarbageCollectorResetEvent)
    {
      DbgPrint("GarbageCollector : Reset received.\n");
    }

    /*
     * One of processes dead.
     * Send unmount explicite.
     */

    else
    {
      DOKAN_CONTROL *dc = handlesData + id;

      DbgPrintW(L"GarbageCollector : [%d][%s][%s] signaled.\n",
                    dc -> clientPid, dc -> MountPoint, dc -> DeviceName);

      dc -> Type = DOKAN_CONTROL_UNMOUNT;

      DokanControl(dc);
    }

    /*
     * Clean handles table before next epoch.
     * Skip last handle, becouse it's ready event.
     */

    for (id = 0; id < count - 1; id++)
    {
      CloseHandle(handles[id]);
    }
  }

  /*
   * Error handler.
   */

  exitCode = 0;

  fail:

  if (exitCode)
  {
    DbgPrint("ERROR: Cannot monitor client processes.\n"
                 "Error code is : %d.\n", GetLastError());

    /*
     * Send STOP message to itself.
     */

    SetEvent(GarbageCollectorStopEvent);
  }
  else
  {
    CloseHandle(GarbageCollectorStopEvent);
  }

  DbgPrint("<- GarbageCollectorHandler()...\n");

  return exitCode;
}

/*
 * Initialize objects needed for garbage collector thread.
 *
 * RETURNS: 0 if OK.
 */

int GarbageCollectorInit()
{
  int exitCode = -1;

  DbgPrint("-> GarbageCollectorInit()...\n");

  GarbageCollectorStop = CreateEvent(NULL, FALSE, FALSE, NULL);

  /*
   * Create reset event to tell garbage collector thread,
   * when new process arrived.
   */

  if (GarbageCollectorResetEvent == NULL)
  {
    DbgPrint("GarbageCollectorInit : Creating reset event...");

    GarbageCollectorResetEvent = CreateEvent(NULL, FALSE, FALSE, NULL);

    FAILEX(GarbageCollectorResetEvent == NULL,
               "ERROR: Cannot create reset event.\n");
  }

  /*
   *  Create garbage collector thread.
   */

  if (GarbageCollectorThread == NULL)
  {
    GarbageCollectorStopEvent = CreateEvent(NULL, FALSE, FALSE, NULL);

    DbgPrint("GarbageCollectorInit : Creating collector thread...");

    GarbageCollectorThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) GarbageCollectorHandler,
                                              NULL, 0, NULL);

    FAILEX(GarbageCollectorThread == NULL,
               "ERROR: Cannot create garbage collector thread.\n");
  }

  /*
   * Error handler.
   */

  exitCode = 0;

  fail:

  if (exitCode)
  {
    DbgPrint("ERROR: Cannot initialize garbage collector.\n"
                 "Error code is : %d.\n", GetLastError());
  }

  DbgPrint("<- GarbageCollectorInit()...\n");

  return exitCode;
}

/*
 * Clean up objects needed for garbage collector thread.
 */

void GarbageCollectorCleanUp()
{
  DbgPrint("-> GarbageCollectorCleanUp()...\n");

  if (GarbageCollectorStop != NULL && GarbageCollectorThread != NULL)
  {
    SetEvent(GarbageCollectorStop);

    WaitForSingleObject(GarbageCollectorThread, 3000);

    CloseHandle(GarbageCollectorStop);
  }

  CloseHandle(GarbageCollectorThread);
  CloseHandle(GarbageCollectorResetEvent);

  DbgPrint("<- GarbageCollectorCleanUp()...\n");
}

#endif /* NX_FIXME */

/*
 * END OF NX FIXME.
 */
